import logging
import requests
import socket
import sys
from builtins import str
from datetime import datetime

import click
from pkginfo import Installed
from six import PY3, BytesIO, StringIO, string_types

from dtale.utils import make_list

logger = logging.getLogger(__name__)

LOG_LEVELS = dict(
    debug=logging.DEBUG,
    info=logging.INFO,
    warning=logging.WARNING,
    error=logging.ERROR,
    critical=logging.CRITICAL,
)


def setup_logging(logfile, log_level, verbose=False, dedicated_logger=False):
    """
    Utility method for setting up logging configuration

    :param logfile: location of D-Tale logs
    :type logfile: str
    :param log_level: D-Tale's implementation of standard logging levels
    :type log_level: str, options are debug, info, warning, error or, critical
    :param verbose: turns on verbose logging, defaults to False
    :type verbose: bool, optional
    :param dedicated_logger: creates a dedicated dtale logger, defaults to False
    :type dedicated_logger: bool, optional
    :return:
    """

    if log_level == "verbose" or verbose:
        log_level = LOG_LEVELS["debug"]
    elif log_level:
        log_level = LOG_LEVELS[log_level]
    else:
        log_level = LOG_LEVELS["info"]

    logger_name = "dtale" if dedicated_logger else None
    logging.getLogger(logger_name).handlers = []

    fmt = "%(asctime)s - %(levelname)-8s - %(message)s"
    try:
        logging.basicConfig(format=fmt, level=log_level)
    except BaseException:
        # #202: when using Pyzo IDE the basicConfig has already been set and if you try to set it again
        # then it will cause a maximum recursion exception
        logging.getLogger(logger_name).setLevel(log_level)

    if logfile:
        fh = logging.FileHandler(logfile, mode="w")
        fh.setFormatter(logging.Formatter(fmt))
        logging.getLogger(logger_name).addHandler(fh)

    for handler in logging.getLogger(logger_name).handlers:
        handler.setLevel(log_level)
        handler.setFormatter(logging.Formatter(fmt))

    logger.debug("{}".format(" ".join(sys.argv)))
    try:
        logger.debug("Hostname: {}".format(socket.gethostname()))
    except Exception as e:
        logger.exception(e)


def loader_options(key, params):
    """
    Builds a decorator which will append a list of click parameters based on
    a combination of prefix & parameter name

    loader_options('foo', ['bar', 'baz']) => click.option('--foo-bar'), click.option('--foo-baz')

    :param key: parameter prefix
    :type key: str
    :param params: parameters names
    :type params: list
    :return: filter
    :rtype: func
    """

    def decorator(f):
        for p in params:
            f = click.option(
                "--" + key + "-" + p, help="Override {} {}".format(key, p)
            )(f)
        return f

    return decorator


def get_loader_options(key, properties, options):
    """
    Filters dictionary of click parameters for ones which start with a certain prefix

    :param key: click option prefix
    :type key: str
    :param properties: sub-properties of the click command key
    :type properties: list of str
    :param options: click options
    :type options: dict
    :return: dictionary of click options with start with key
    :rtype: dict
    """

    def _build_key(option):
        segs = option.split("_")
        if len(segs) == 1:
            return ""
        return option.split("{}_".format(key))[-1]

    final_key = "_".join(key.split("-"))
    selected_opts = []
    if not len(make_list(properties)):
        selected_opts = [final_key]
    else:
        for prop in properties:
            p_names = make_list(prop["name"] if isinstance(prop, dict) else prop)
            for p_name in p_names:
                selected_opts.append("{}_{}".format(final_key, p_name))

    return dict(
        (
            (_build_key(k), v)
            for k, v in options.items()
            if k in selected_opts
            if v is not None
        )
    )


def get_log_options(options):
    """
    Click logging options (logfile, log_level, verbose)
    """
    names = ["logfile", "log_level", "verbose"]
    return get_named_options(names, options)


def get_named_options(names, options):
    """
    Filters click options for options in a specific subset of names
    """
    return {k: v for k, v in options.items() if k in names}


def retrieve_version(name):
    """
    Retrieves versioning of specific package name
    """
    try:
        return Installed(name).version
    except BaseException:
        return "unknown"


def get_args(click_wrapper):
    """
    Retrieves arguments being passed in from the command-line and applying them to a click command

    :param click_wrapper: click context
    :return: list of arguments
    :rtype: list
    """

    # Get the options defined on cli
    params = []
    for opt in click_wrapper.params:
        params.extend(opt.opts)

    args = sys.argv[1:]

    # Reorder the args so that cli options come before
    # subcommand options. This is a click limitation.
    head, tail = [], []
    for a in args:
        if a in params or head and head[-1] in params:
            head.append(a)
        else:
            tail.append(a)
    return head + tail


def run(click_wrapper):
    """
    Wrapper function for executing click commands and handling exceptions produced from
    click commands with versioning boilerplate

    :param click_wrapper: click context
    """

    t1 = datetime.now()
    try:
        try:
            dtale_name = "dtale"
            ver_info = retrieve_version(dtale_name)
            logger.debug("{} ver: {}".format(dtale_name, ver_info))
        except BaseException:
            logger.debug("failure to retrieve metadata for: {}".format(dtale_name))

        args = get_args(click_wrapper)
        click_wrapper(args)
        sys.exit(0)
    except Exception as ex:
        logger.exception("Fatal error: " + str(ex))
        sys.exit(1)
    finally:
        logger.info("Elapsed time: %s" % (datetime.now() - t1))


def loader_prop_keys(prop_cfgs):
    return [p if isinstance(p, string_types) else p["name"] for p in prop_cfgs]


def handle_str_content(resp):
    return BytesIO(resp.content) if PY3 else StringIO(resp.content.decode("utf-8"))


def is_url(path):
    return path.startswith("http://") or path.startswith("https://")


def handle_path(path, kwargs, resp_handler=handle_str_content):
    if is_url(path):
        proxy = kwargs.pop("proxy", None)
        req_kwargs = {}
        if proxy is not None:
            req_kwargs["proxies"] = dict(http=proxy, https=proxy)
        resp = requests.get(path, **req_kwargs)
        assert resp.status_code == 200
        return resp_handler(resp)
    return path



import sys

from threading import Lock
from werkzeug.wsgi import get_path_info

_default_encoding = sys.getdefaultencoding()


def _to_str(x, charset=_default_encoding, errors="strict", allow_none_charset=False):
    if x is None or isinstance(x, str):
        return x

    if not isinstance(x, (bytes, bytearray)):
        return str(x)

    if charset is None:
        if allow_none_charset:
            return x

    return x.decode(charset, errors)


def pop_path_info(environ, charset="utf-8", errors="replace"):
    """Removes and returns the next segment of `PATH_INFO`, pushing it onto
    `SCRIPT_NAME`.  Returns `None` if there is nothing left on `PATH_INFO`.
    If the `charset` is set to `None` bytes are returned.
    If there are empty segments (``'/foo//bar``) these are ignored but
    properly pushed to the `SCRIPT_NAME`:
    >>> env = {'SCRIPT_NAME': '/foo', 'PATH_INFO': '/a/b'}
    >>> pop_path_info(env)
    'a'
    >>> env['SCRIPT_NAME']
    '/foo/a'
    >>> pop_path_info(env)
    'b'
    >>> env['SCRIPT_NAME']
    '/foo/a/b'
    .. deprecated:: 2.2
        Will be removed in Werkzeug 2.3.
    .. versionadded:: 0.5
    .. versionchanged:: 0.9
       The path is now decoded and a charset and encoding
       parameter can be provided.
    :param environ: the WSGI environment that is modified.
    :param charset: The ``encoding`` parameter passed to
        :func:`bytes.decode`.
    :param errors: The ``errors`` paramater passed to
        :func:`bytes.decode`.
    """

    path = get_path_info(environ)
    if not path:
        return None

    script_name = environ.get("SCRIPT_NAME", "")

    # shift multiple leading slashes over
    old_path = path
    path = path.lstrip("/")
    if path != old_path:
        script_name += "/" * (len(old_path) - len(path))

    if "/" not in path:
        environ["PATH_INFO"] = ""
        environ["SCRIPT_NAME"] = script_name + path
        rv = path.encode("latin1")
    else:
        segment, path = path.split("/", 1)
        environ["PATH_INFO"] = "/{}".format(path)
        environ["SCRIPT_NAME"] = script_name + segment
        rv = segment.encode("latin1")

    return _to_str(rv, charset, errors, allow_none_charset=True)  # type: ignore


def peek_path_info(environ, charset="utf-8", errors="replace"):
    """Returns the next segment on the `PATH_INFO` or `None` if there
    is none.  Works like :func:`pop_path_info` without modifying the
    environment:
    >>> env = {'SCRIPT_NAME': '/foo', 'PATH_INFO': '/a/b'}
    >>> peek_path_info(env)
    'a'
    >>> peek_path_info(env)
    'a'
    If the `charset` is set to `None` bytes are returned.
    .. deprecated:: 2.2
        Will be removed in Werkzeug 2.3.
    .. versionadded:: 0.5
    .. versionchanged:: 0.9
       The path is now decoded and a charset and encoding
       parameter can be provided.
    :param environ: the WSGI environment that is checked.
    """

    segments = get_path_info(environ).lstrip("/").split("/", 1)
    if segments:
        return _to_str(  # type: ignore
            segments[0].encode("latin1"), charset, errors, allow_none_charset=True
        )
    return None


class PathDispatcher(object):
    def __init__(self, default_app, create_app):
        self.default_app = default_app
        self.create_app = create_app
        self.lock = Lock()
        self.instances = {}

    def get_application(self, prefix):
        with self.lock:
            app = self.instances.get(prefix)
            if app is None:
                app = self.create_app(prefix)
                if app is not None:
                    self.instances[prefix] = app
            return app

    def __call__(self, environ, start_response):
        # TODO: update to use get_path_info
        app = self.get_application(peek_path_info(environ))
        if app is not None:
            pop_path_info(environ)
        else:
            app = self.default_app
        return app(environ, start_response)



import string
import inspect

from logging import getLogger
from six import PY3

from dtale.utils import dict_merge, format_data

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping

logger = getLogger(__name__)

APP_SETTINGS = {
    "theme": "light",
    "pin_menu": False,
    "language": "en",
    "github_fork": False,
    "hide_shutdown": False,
    "max_column_width": None,
    "max_row_height": None,
    "main_title": None,
    "main_title_font": None,
    "query_engine": "python",
    "open_custom_filter_on_startup": False,
    "open_predefined_filters_on_startup": False,
    "hide_drop_rows": False,
    "hide_header_editor": False,
    "lock_header_menu": False,
    "hide_header_menu": False,
    "hide_main_menu": False,
    "hide_column_menus": False,
    "enable_custom_filters": False,
    "enable_web_uploads": False,
    "hide_row_expanders": False,
}

AUTH_SETTINGS = {"active": False, "username": None, "password": None}

CHART_SETTINGS = {"scatter_points": 15000, "3d_points": 40000}


class DtaleInstance(object):
    _dataset = None
    _dataset_dim = None
    _dtypes = None
    _metadata = None
    _context_variables = None
    _history = None
    _settings = None
    _name = ""
    _rows = 0

    def __init__(self, data):
        self._data = data
        self._rows = 0 if self._data is None else len(data)

    def load_data(self):
        return self._data

    def rows(self, **kwargs):
        return self._rows

    @property
    def is_large(self):
        return False

    @property
    def data(self):
        return self.load_data()

    @property
    def name(self):
        return self._name

    @property
    def dataset(self):
        return self._dataset

    @property
    def dataset_dim(self):
        return self._dataset_dim

    @property
    def dtypes(self):
        return self._dtypes

    @property
    def metadata(self):
        return self._metadata

    @property
    def context_variables(self):
        return self._context_variables

    @property
    def history(self):
        return self._history

    @property
    def settings(self):
        return self._settings

    @property
    def is_xarray_dataset(self):
        if self._dataset is not None:
            return True
        return False

    @data.setter
    def data(self, data):
        self._data = data

    @name.setter
    def name(self, name):
        self._name = name

    @dataset.setter
    def dataset(self, dataset):
        self._dataset = dataset

    @dataset_dim.setter
    def dataset_dim(self, dataset_dim):
        self._dataset_dim = dataset_dim

    @dtypes.setter
    def dtypes(self, dtypes):
        self._dtypes = dtypes

    @context_variables.setter
    def context_variables(self, context_variables):
        self._context_variables = context_variables

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = metadata

    @history.setter
    def history(self, history):
        self._history = history

    @settings.setter
    def settings(self, settings):
        self._settings = settings


LARGE_ARCTICDB = 1000000


def get_num_rows(lib, symbol):
    try:
        return lib._nvs.get_num_rows(symbol)
    except BaseException:
        read_options = lib._nvs._get_read_options()
        version_query = lib._nvs._get_version_query(None)
        dit = lib._nvs.version_store.read_descriptor(
            symbol, version_query, read_options
        )
        return dit.timeseries_descriptor.total_rows


class DtaleArcticDBInstance(DtaleInstance):
    def __init__(self, data, data_id, parent):
        super(DtaleArcticDBInstance, self).__init__(data)
        self.parent = parent
        data_id_segs = (data_id or "").split("|")
        symbol = data_id_segs[-1]
        if len(data_id_segs) > 1:
            self.lib_name = data_id_segs[0]
            if not parent.lib or self.lib_name != parent.lib.name:
                parent.update_library(self.lib_name)
        else:
            self.lib_name = self.parent.lib.name if self.parent.lib else None

        lib = self.parent.get_library(self.lib_name) if self.lib_name else None
        self.symbol = symbol
        self._rows = 0
        self._cols = 0
        self._base_df = None
        if lib and self.symbol and self.symbol in self.parent._symbols[self.lib_name]:
            self._rows = get_num_rows(lib, self.symbol)
            self._base_df = self.load_data(row_range=[0, 1])
            self._cols = len(format_data(self._base_df)[0].columns)
        elif (
            lib
            and self.symbol
            and self.symbol not in self.parent._symbols[self.lib_name]
        ):
            raise ValueError(
                "Symbol ({}) not in library, {}! Please select another symbol.".format(
                    symbol, self.lib_name
                )
            )

    def load_data(self, **kwargs):
        from arcticdb.version_store._store import VersionedItem

        if self.symbol not in self.parent._symbols[self.lib_name]:
            raise ValueError(
                "{} does not exist in {}!".format(self.symbol, self.lib_name)
            )

        data = self.parent.get_library(self.lib_name)._nvs.read(self.symbol, **kwargs)
        if isinstance(data, VersionedItem):
            return data.data
        return data

    def rows(self, **kwargs):
        if kwargs.get("query_builder"):
            version_query, read_options, read_query = self.parent.get_library(
                self.lib_name
            )._nvs._get_queries(
                None, None, None, None, query_builder=kwargs["query_builder"]
            )
            read_result = self.parent.get_library(self.lib_name)._nvs._read_dataframe(
                self.symbol, version_query, read_query, read_options
            )
            return len(read_result.frame_data.value.data[0])
        return self._rows

    @property
    def base_df(self):
        return self._base_df

    @property
    def is_large(self):
        if self.rows() > LARGE_ARCTICDB:
            return True
        if self._cols > 50:
            return True
        return False

    @property
    def data(self):
        return self.load_data()

    @data.setter
    def data(self, data):
        try:
            self.parent.get_library(self.lib_name).write(self.symbol, data)
        except BaseException:
            pass


class DtaleBaseStore(dict):
    def build_instance(self, data_id, data=None):
        return DtaleInstance(data)


class DtaleArcticDB(DtaleBaseStore):
    """Interface allowing dtale to use 'arcticdb' databases for global data storage."""

    def __init__(self, uri=None, library=None, **kwargs):
        from arcticdb import Arctic

        self._db = dict()
        self.uri = uri
        self.conn = Arctic(self.uri)
        self.lib = None
        self._libraries = []
        self._symbols = {}
        self.load_libraries()
        self.update_library(library)

    def get_library(self, library_name):
        return self.conn[library_name]

    def update_library(self, library=None):
        if self.lib and library == self.lib.name:  # library already selected
            return
        if library in self._libraries:
            self.lib = self.get_library(library)
            if library not in self._symbols:
                self.load_symbols()
        elif library is not None:
            raise ValueError("Library '{}' does not exist!".format(library))

    def load_libraries(self):
        self._libraries = sorted(self.conn.list_libraries())

    @property
    def libraries(self):
        return self._libraries

    def load_symbols(self, library=None):
        self._symbols[library or self.lib.name] = sorted(
            (self.conn[library] if library else self.lib).list_symbols()
        )

    @property
    def symbols(self):
        return self._symbols[self.lib.name]

    def build_instance(self, data_id, data=None):
        if data_id is None:
            return DtaleInstance(data)
        return DtaleArcticDBInstance(data, data_id, self)

    def get(self, key, **kwargs):
        if key is None:
            return self.build_instance(key)
        key = str(key)
        if key not in self._db:
            self._db[key] = self.build_instance(key)
        return self._db[key]

    def __setitem__(self, key, value):
        if key is None:
            return
        key = str(key)
        self._db[key] = value

    def __delitem__(self, key):
        if key is None:
            return
        key = str(key)
        # TODO: should we actually delete from ArcticDB???
        del self._db[key]

    def __contains__(self, key):
        key = str(key)
        return key in self._db

    def clear(self):
        pass

    def to_dict(self):
        return dict(self._db)

    def items(self):
        return self.to_dict().items()

    def keys(self):
        return self.to_dict().keys()

    def __len__(self):
        return len(self.keys())

    def save_db(self):
        raise NotImplementedError


class DefaultStore(object):
    def __init__(self):
        self._data_store = DtaleBaseStore()
        self._data_names = dict()

    # Use int for data_id for easier sorting
    def build_data_id(self):
        if len(self._data_store) == 0:
            return "1"

        def parse_int(x):
            try:
                return int(x)
            except ValueError:
                return None

        ids = list(filter(None, map(parse_int, self._data_store.keys())))
        if not len(ids):
            return "1"
        return str(max(ids) + 1)

    @property
    def is_arcticdb(self):
        return isinstance(self._data_store, DtaleArcticDB)

    # exposing  _data_store for custom data store plugins.
    @property
    def store(self):
        return self._data_store

    @store.setter
    def store(self, new_store):
        self._data_store = new_store

    def keys(self):
        return list(self._data_store.keys())

    def items(self):
        return self._data_store.items()

    def contains(self, key):
        if key is None:
            return False
        return str(key) in self._data_store

    # this should be a property but somehow it stays 0 no matter what.
    def size(self):
        return len(self._data_store)

    def get_data_inst(self, data_id):
        # handle non-exist data_id
        if data_id is None:
            return self._data_store.build_instance(data_id)

        if str(data_id) not in self._data_store:
            self._data_store[str(data_id)] = self._data_store.build_instance(data_id)

        return self._data_store.get(str(data_id))

    def new_data_inst(self, data_id=None, instance=None):
        if data_id is None:
            data_id = self.build_data_id()
        data_id = str(data_id)
        new_data = instance or self._data_store.build_instance(data_id)
        self._data_store[data_id] = new_data
        return data_id

    def get_data(self, data_id, **kwargs):
        return self.get_data_inst(data_id).load_data(**kwargs)

    def get_data_id_by_name(self, data_name):
        data_id = next(
            (
                value
                for key, value in self._data_names.items()
                if convert_name_to_url_path(key) == data_name or key == data_name
            ),
            None,
        )
        return data_id

    def get_dataset(self, data_id):
        return self.get_data_inst(data_id).dataset

    def get_dataset_dim(self, data_id):
        return self.get_data_inst(data_id).dataset_dim

    def get_dtypes(self, data_id):
        return self.get_data_inst(data_id).dtypes

    def get_context_variables(self, data_id):
        return self.get_data_inst(data_id).context_variables

    def get_history(self, data_id):
        return self.get_data_inst(data_id).history

    def get_name(self, data_id):
        return self.get_data_inst(data_id).name

    def get_settings(self, data_id):
        return self.get_data_inst(data_id).settings

    def get_query(self, data_id):
        if load_flag(data_id, "enable_custom_filters", False):
            curr_settings = self.get_settings(data_id) or {}
            return curr_settings.get("query")
        return None

    def get_metadata(self, data_id):
        return self.get_data_inst(data_id).metadata

    def set_data(self, data_id=None, val=None):
        if data_id is None:
            data_id = self.new_data_inst()
        data_id = str(data_id)
        if data_id not in self._data_store.keys():
            data_id = self.new_data_inst(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.data = val
        self._data_store[data_id] = data_inst

    def set_dataset(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.dataset = val
        self._data_store[data_id] = data_inst

    def set_dataset_dim(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.dataset_dim = val
        self._data_store[data_id] = data_inst

    def set_dtypes(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.dtypes = val
        self._data_store[data_id] = data_inst

    def set_name(self, data_id, val):
        if val in [None, ""]:
            return
        if val in self._data_names:
            raise Exception("Name {} already exists!".format(val))
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        self._data_names[val] = data_id
        data_inst.name = val
        self._data_store[data_id] = data_inst

    def set_context_variables(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.context_variables = val
        self._data_store[data_id] = data_inst

    def set_settings(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.settings = val
        self._data_store[data_id] = data_inst

    def set_metadata(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.metadata = val
        self._data_store[data_id] = data_inst

    def set_history(self, data_id, val):
        data_id = str(data_id)
        data_inst = self.get_data_inst(data_id)
        data_inst.history = val
        self._data_store[data_id] = data_inst

    def delete_instance(self, data_id):
        data_id = str(data_id)
        instance = self._data_store.get(data_id)
        if instance:
            if instance.name:
                try:
                    del self._data_names[instance.name]
                except KeyError:
                    pass
            try:
                del self._data_store[data_id]
            except KeyError:
                pass

    def clear_store(self):
        self._data_store.clear()
        self._data_names.clear()


"""
This block dynamically exports functions from DefaultStore class.
It's here for backward compatibility reasons.
It may trigger linter errors in other py files because functions are not statically exported.
"""
_default_store = DefaultStore()
fn_list = list(
    filter(
        lambda x: not x.startswith("_"),
        [x[0] for x in inspect.getmembers(DefaultStore)],
    )
)

for fn_name in fn_list:
    globals()[fn_name] = getattr(_default_store, fn_name)


# for tests. default_store is always initialized.
def use_default_store():
    new_store = DtaleBaseStore()
    for k, v in _as_dict(_default_store.store).items():
        new_store[str(k)] = v
    _default_store.store.clear()
    _default_store.store = new_store
    globals()["is_arcticdb"] = getattr(_default_store, "is_arcticdb")
    globals()["store"] = getattr(_default_store, "store")
    pass


def drop_punctuation(val):
    if PY3:
        return val.translate(str.maketrans(dict.fromkeys(string.punctuation)))
    return val.translate(string.maketrans("", ""), string.punctuation)


def convert_name_to_url_path(name):
    if name is None:
        return None
    url_name = drop_punctuation("{}".format(name))
    url_name = url_name.lower()
    return "_".join(url_name.split(" "))


def get_dtype_info(data_id, col):
    dtypes = get_dtypes(data_id)  # noqa: F821
    return next((c for c in dtypes or [] if c["name"] == col), None)


def update_settings(data_id, settings):
    curr_settings = _default_store.get_settings(data_id) or {}
    updated_settings = dict_merge(curr_settings, settings)
    _default_store.set_settings(data_id, updated_settings)


def get_app_settings():
    return APP_SETTINGS


def set_app_settings(settings):
    for prop, val in settings.items():
        APP_SETTINGS[prop] = val

    instance_updates = {}
    if settings.get("hide_shutdown") is not None:
        instance_updates["hide_shutdown"] = settings.get("hide_shutdown")
    if settings.get("hide_header_editor") is not None:
        instance_updates["hide_header_editor"] = settings.get("hide_header_editor")
    if settings.get("lock_header_menu") is not None:
        instance_updates["lock_header_menu"] = settings.get("lock_header_menu")
    if settings.get("hide_header_menu") is not None:
        instance_updates["hide_header_menu"] = settings.get("hide_header_menu")
    if settings.get("hide_main_menu") is not None:
        instance_updates["hide_main_menu"] = settings.get("hide_main_menu")
    if settings.get("hide_column_menus") is not None:
        instance_updates["hide_column_menus"] = settings.get("hide_column_menus")
    if settings.get("hide_row_expanders") is not None:
        instance_updates["hide_row_expanders"] = settings.get("hide_row_expanders")
    if settings.get("theme") is not None:
        instance_updates["theme"] = settings.get("theme")
    if settings.get("enable_custom_filters") is not None:
        instance_updates["enable_custom_filters"] = settings.get(
            "enable_custom_filters"
        )
        if instance_updates["enable_custom_filters"]:
            logger.warning(
                (
                    "Turning on custom filtering. Custom filters are vulnerable to code injection attacks, please only "
                    "use in trusted environments."
                )
            )
    if settings.get("enable_web_uploads") is not None:
        instance_updates["enable_web_uploads"] = settings.get("enable_web_uploads")
        if instance_updates["enable_web_uploads"]:
            logger.warning(
                (
                    "Turning on Web uploads. Web uploads are vulnerable to blind server side request forgery, please "
                    "only use in trusted environments."
                )
            )

    if _default_store.size() > 0 and len(instance_updates):
        for data_id in _default_store.keys():
            update_settings(data_id, instance_updates)


def get_auth_settings():
    return AUTH_SETTINGS


def set_auth_settings(settings):
    for prop, val in settings.items():
        AUTH_SETTINGS[prop] = val


def get_chart_settings():
    return CHART_SETTINGS


def set_chart_settings(settings):
    for prop, val in settings.items():
        CHART_SETTINGS[prop] = val


def cleanup(data_id=None):
    if data_id is None:
        _default_store.clear_store()
    else:
        _default_store.delete_instance(data_id)


def update_id(old_data_id, new_data_id):
    if _default_store.contains(new_data_id):
        raise Exception("Data already exists for id ({})".format(new_data_id))
    curr_data = _default_store.get_data_inst(old_data_id)
    _default_store.delete_instance(old_data_id)
    data_id = str(new_data_id)
    _default_store.new_data_inst(data_id, curr_data)
    return new_data_id


def load_flag(data_id, flag_name, default):
    import dtale

    app_settings = get_app_settings()
    curr_settings = get_settings(data_id) or {}  # noqa: F821
    global_flag = getattr(dtale, flag_name.upper(), default)
    if global_flag != default:
        return global_flag
    if flag_name in app_settings and app_settings[flag_name] != default:
        return app_settings[flag_name]
    return curr_settings.get(flag_name, app_settings.get(flag_name, default))


def _as_dict(store):
    """Return the dict representation of a data store.
    Stores must either be an instance of MutableMapping OR have a to_dict method.

    :param store: data store (dict, redis connection, etc.)
    :return: dict
    """
    return dict(store.items()) if isinstance(store, MutableMapping) else store.to_dict()


def use_store(store_class, create_store):
    """
    Customize how dtale stores and retrieves global data.
    By default it uses global dictionaries, but this can be problematic if
    there are memory limitations or multiple python processes are running.
    Ex: a web server with multiple workers (processes) for processing requests.

    :param store_class: Class providing an interface to the data store. To be valid, it must:
                        1. Implement get, keys, items clear, __setitem__, __delitem__, __iter__, __len__, __contains__.
                        2. Either be a subclass of MutableMapping or implement the 'to_dict' method.
    :param create_store: Factory function for producing instances of <store_class>.
                         Must take 'name' as the only parameter.
    :return: None
    """

    assert inspect.isclass(store_class), "Must be a class"
    assert all(
        hasattr(store_class, a)
        for a in (
            "get",
            "clear",
            "keys",
            "items",
            "__setitem__",
            "__delitem__",
            "__len__",
            "__contains__",
        )
    ), "Missing required methods"
    assert issubclass(store_class, MutableMapping) or hasattr(
        store_class, "to_dict"
    ), 'Must subclass MutableMapping or implement "to_dict"'

    assert inspect.isfunction(create_store), "Must be a function"
    if PY3:
        assert list(inspect.signature(create_store).parameters) == [
            "name"
        ], 'Must take "name" as the only parameter'
    else:
        assert inspect.getargspec(create_store).args == [
            "name"
        ], 'Must take "name" as the only parameter'

    def convert(old_store, name):
        """Convert a data store to the new type
        :param old_store: old data store
        :param name: name associated with this data store
        :return: new data store
        """
        new_store = create_store(name)
        assert isinstance(new_store, store_class)
        new_store.clear()
        for k, v in _as_dict(old_store).items():
            new_store[k] = v
        old_store.clear()
        return new_store

    _default_store.store = convert(_default_store.store, "default_store")
    globals()["is_arcticdb"] = getattr(_default_store, "is_arcticdb")
    globals()["store"] = getattr(_default_store, "store")


def use_shelve_store(directory):
    """
    Configure dtale to use python's standard 'shelve' library for a persistent global data store.

    :param directory: directory that the shelve db files will be stored in
    :type directory: str
    :return: None
    """
    import shelve
    import time
    from os.path import join
    from threading import Thread

    class DtaleShelf(DtaleBaseStore):
        """Interface allowing dtale to use 'shelf' databases for global data storage."""

        def __init__(self, filename):
            self.filename = filename
            self.db = shelve.open(self.filename, flag="c", writeback=True)
            # super hacky autosave
            t = Thread(target=self.save_db)
            t.daemon = True
            t.start()

        def get(self, key):
            # using str here because shelve doesn't support int keys
            key = str(key)
            return self.db.get(key)

        def __setitem__(self, key, value):
            key = str(key)
            self.db[key] = value
            self.db.sync()

        def __delitem__(self, key):
            key = str(key)
            del self.db[key]
            self.db.sync()

        def __contains__(self, key):
            key = str(key)
            return key in self.db

        def clear(self):
            self.db.clear()
            self.db.sync()

        def to_dict(self):
            return dict(self.db)

        def items(self):
            return self.to_dict().items()

        def keys(self):
            return self.to_dict().keys()

        def __len__(self):
            return len(self.db)

        def save_db(self):
            while True:
                self.db.sync()
                time.sleep(5)

    def create_shelf(name):
        file_path = join(directory, name)
        return DtaleShelf(file_path)

    use_store(DtaleShelf, create_shelf)


def use_redis_store(directory, *args, **kwargs):
    """Configure dtale to use redis for the global data store. Useful for web servers.

    :param db_folder: folder that db files will be stored in
    :type db_folder: str
    :param args: All other arguments supported by the redislite.Redis() class
    :param kwargs: All other keyword arguments supported by the redislite.Redis() class
    :return: None
    """
    import pickle
    from os.path import join

    try:
        from redislite import Redis
    except ImportError:
        raise Exception("redislite must be installed")

    # redis-py 7.2.0 accesses conn.host/conn.port for observability on all
    # connection types, but UnixDomainSocketConnection only sets 'path'.
    # Patch in fallbacks so redislite (which uses Unix sockets) doesn't break.
    from redis.connection import UnixDomainSocketConnection as _UDS

    _orig_uds_init = _UDS.__init__

    def _patched_uds_init(self, *a, **kw):
        _orig_uds_init(self, *a, **kw)
        if not hasattr(self, "port"):
            self.port = 0
        if not hasattr(self, "host"):
            self.host = self.path

    _UDS.__init__ = _patched_uds_init

    class DtaleRedis(DtaleBaseStore, Redis):
        """Wrapper class around Redis() to make it work as a global data store in dtale."""

        def __init__(self, file_path, *args, **kwargs):
            super(Redis, self).__init__(file_path, *args, **kwargs)

        def __setitem__(self, key, value):
            key = str(key)
            self.set(key, value)

        def __delitem__(self, key):
            key = str(key)
            super(Redis, self).__delitem__(str(key))

        def __contains__(self, key):
            key = str(key)
            return super(Redis, self).__contains__(str(key))

        def get(self, name, *args, **kwargs):
            value = super(Redis, self).get(name, *args, **kwargs)
            if value is not None:
                return pickle.loads(value)

        def keys(self):
            return [str(k) for k in super(Redis, self).keys()]

        def set(self, name, value, *args, **kwargs):
            value = pickle.dumps(value)
            return super(Redis, self).set(name, value, *args, **kwargs)

        def clear(self):
            self.flushdb()

        def to_dict(self):
            return {k.decode("utf-8"): self.get(k) for k in super(Redis, self).keys()}

        def items(self):
            return self.to_dict().items()

        def __len__(self):
            return len(self.keys())

    def create_redis(name):
        file_path = join(directory, name + ".db")
        return DtaleRedis(file_path, *args, **kwargs)

    use_store(DtaleRedis, create_redis)


def use_arcticdb_store(*args, **kwargs):
    """
    Configure dtale to use arcticdb for a persistent global data store.

    :param uri: URI that arcticdb will connect to (local file storage: lmdb:///<path>)
    :type uri: str
    :param library: default library to load from arcticdb URI
    :type library: str
    :param symbol: defualt symbol to load from library
    :type symbol: str
    :return: None
    """

    def create_arcticdb(name):
        return DtaleArcticDB(**kwargs)

    use_store(DtaleArcticDB, create_arcticdb)



# AUTO GENERATED FILE - DO NOT EDIT

from dash.development.base_component import Component, _explicitize_args


class DashColorscales(Component):
    """A DashColorscales component.
    DashColorscales is a Dash wrapper for `react-colorscales`.
    It takes an array of colors, `colorscale`, and
    displays a UI for modifying it or choosing a new scale.

    Keyword arguments:

    - id (string; optional):
        The ID used to identify this compnent in Dash callbacks.

    - colorscale (list; optional):
        Optional: Initial colorscale to display. Default is Viridis.

    - fixSwatches (boolean; optional):
        Optional: Set to `True` to fix the number of colors in the scale.

    - nSwatches (number; optional):
        Optional: Initial number of colors in scale to display."""

    _children_props = []
    _base_nodes = ["children"]
    _namespace = "components"
    _type = "DashColorscales"

    @_explicitize_args
    def __init__(
        self,
        id=Component.REQUIRED,
        colorscale=Component.UNDEFINED,
        nSwatches=Component.UNDEFINED,
        fixSwatches=Component.UNDEFINED,
        **kwargs
    ):
        self._prop_names = ["id", "colorscale", "fixSwatches", "nSwatches"]
        self._valid_wildcard_attributes = []
        self.available_properties = ["id", "colorscale", "fixSwatches", "nSwatches"]
        self.available_wildcard_properties = []
        _explicit_args = kwargs.pop("_explicit_args")
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(DashColorscales, self).__init__(**args)



import copy

import numpy as np
import pandas as pd
from pandas.tseries.offsets import Day

import dtale.global_state as global_state
import dtale.pandas_util as pandas_util
from dtale.column_analysis import handle_cleaners
from dtale.constants import CHART_JOINER_CHAR
from dtale.query import build_col_key, run_query
from dtale.utils import (
    ChartBuildingError,
    build_formatters,
    classify_type,
    find_dtype,
    find_dtype_formatter,
    flatten_lists,
    get_dtypes,
    json_int,
    make_list,
    triple_quote,
)

YAXIS_CHARTS = ["line", "bar", "scatter"]
ZAXIS_CHARTS = ["heatmap", "3d_scatter", "surface"]
NON_EXT_AGGREGATION = ZAXIS_CHARTS + ["treemap", "maps"]
ANIMATION_CHARTS = ["line"]
ANIMATE_BY_CHARTS = ["bar", "3d_scatter", "heatmap", "maps", "histogram"]
MAX_GROUPS = 30
MAPBOX_TOKEN = None
AGGS = dict(
    raw="No Aggregation",
    count="Count",
    nunique="Unique Count",
    sum="Sum",
    mean="Mean",
    rolling="Rolling",
    corr="Correlation",
    first="Keep First",
    last="Keep Last",
    drop_duplicates="Remove Duplicates",
    median="Median",
    min="Minimum",
    max="Maximum",
    std="Standard Deviation",
    var="Variance",
    mad="Mean Absolute Deviation",
    prod="Product of All Items",
    pctct="Count (Percentage)",
    pctsum="Percentage Sum",
)
INDEX_COL = "__index__"

CHART_POINTS_LIMIT = (
    "In order to adjust the limitation on the amount of points in charts please update your startup code that "
    "loads the dataframe to D-Tale and add these lines of code before calling 'dtale.show':\n\n"
    "import dtale.global_state as global_state\n\n"
    "global_state.set_chart_settings({'scatter_points': 15000, '3d_points': 40000})\n\n"
    "You could also add the following properties to your global configuration file:\n\n"
    "[charts]\n"
    "scatter_points = 20000\n\n"
    "Documentation on global configurations can be found here:\n"
    "https://github.com/man-group/dtale/blob/master/docs/CONFIGURATION.md"
)


def get_mapbox_token():
    return MAPBOX_TOKEN


def set_mapbox_token(token):
    global MAPBOX_TOKEN

    MAPBOX_TOKEN = token


def valid_chart(chart_type=None, x=None, y=None, z=None, **inputs):
    """
    Helper function to determine based on inputs (chart_type, x, y, z...) whether a chart can be build or not.
    For example, charts must have an x & y value and for 3-dimensional charts they must have a Z-Axis as well.

    :param chart_type: type of chart to build (line, bar, scatter, pie, heatmap, wordcloud, 3dscatter, surface)
    :type chart_type: str, optional
    :param x: column to use for the X-Axis
    :type x: str, optional
    :param y: columns to use for the Y-Axes
    :type y: list of str, optional
    :param z: column to use for the Z-Axis
    :type z: str, optional
    :param inputs: keyword arguments containing
    :return: `True` if executed from test, `False` otherwise
    :rtype: bool
    """
    if chart_type == "maps":
        map_type = inputs.get("map_type")
        if map_type == "choropleth" and all(
            inputs.get(p) is not None for p in ["loc_mode", "loc", "map_val"]
        ):
            if inputs.get("loc_mode") == "geojson-id" and any(
                inputs.get(p) is None for p in ["geojson", "featureidkey"]
            ):
                return False
            return True
        elif map_type in ["scattergeo", "mapbox"] and all(
            inputs.get(p) is not None for p in ["lat", "lon"]
        ):
            return True
        return False

    if chart_type == "candlestick":
        cs_props = ["cs_x", "cs_open", "cs_close", "cs_high", "cs_low"]
        return all(inputs.get(p) is not None for p in cs_props)

    if chart_type in ["treemap", "funnel"]:
        chart_props = ["{}_value", "{}_label"]
        return all(inputs.get(p.format(chart_type)) is not None for p in chart_props)

    if chart_type == "clustergram":
        if (
            inputs.get("clustergram_value") is None
            or inputs.get("clustergram_label") is None
        ):
            return False
        if len(make_list(inputs["clustergram_value"])) == 0:
            return False
        return True

    if chart_type == "pareto":
        pareto_props = ["pareto_x", "pareto_bars", "pareto_line"]
        return all(inputs.get(p) is not None for p in pareto_props)

    if chart_type == "histogram":
        return inputs.get("histogram_col") is not None

    if not x:
        return False

    if not y:
        return chart_type == "wordcloud" and inputs.get("agg") == "count"

    if chart_type in ZAXIS_CHARTS and z is None:
        return False

    if (
        len(inputs.get("extended_aggregation", [])) == 0
        and inputs.get("agg") == "rolling"
        and (inputs.get("window") is None or inputs.get("rolling_comp") is None)
    ):
        return False
    return True


def date_freq_handler(df):
    """
    This returns a column definition handler which returns a series based on the specs from the front-end.
    Column definitions can be a column name 'Col1' or a column name with a frequency 'Col1|M' for
    columns which are of type datetime.

    :Example:
        Col1 -> returns series for Col1
        Col1|M -> returns series for Col1 in monthly format with name 'Col1|M'

    :param df: dataframe whose data needs to be checked
    :type df: :class:`pandas:pandas.DataFrame`
    :return: handler function
    :rtype: func
    """
    dtypes = get_dtypes(df)
    orig_idx = df.index

    def _handler(col_def):
        col_def_segs = col_def.split(CHART_JOINER_CHAR)
        if len(col_def_segs) > 1 and classify_type(dtypes[col_def_segs[0]]) == "D":
            col, freq = col_def_segs
            if freq == "WD":
                code = "df.set_index('{col}').index.dayofweek.values"
                freq_grp = df.set_index(col).index.dayofweek.values
            elif freq == "H2":
                code = "df.set_index('{col}').index.hour.values"
                freq_grp = df.set_index(col).index.hour.values
            else:
                code = "df.set_index('{col}').index.to_period('{freq}').to_timestamp(how='end').values"
                freq_grp = (
                    df.set_index(col)
                    .index.to_period(freq)
                    .to_timestamp(how="end")
                    .values
                )
            code = "\tpd.Series(" + code + ", index=df.index, name='{col_def}'),"
            freq_grp = pd.Series(freq_grp, index=orig_idx, name=col_def)
            return freq_grp, code.format(col=col, freq=freq, col_def=col_def)
        else:
            return df[col_def], "\tdf['{col_def}'],".format(col_def=col_def)

    return _handler


def convert_date_val_to_date(group_val):
    if isinstance(group_val, int):
        return pd.Timestamp(group_val, unit="ms")
    return pd.Timestamp(group_val)


def group_filter_handler(col_def, group_val, group_classifier):
    col_def_segs = col_def.split(CHART_JOINER_CHAR)
    if len(col_def_segs) > 1:
        col, freq = col_def_segs
        if group_val == "NaN":
            return (
                "{col} != {col}".format(col=build_col_key(col)),
                "{}: NaN".format(col),
            )
        if freq == "WD":
            return (
                "{}.dt.dayofweek == {}".format(build_col_key(col), group_val),
                "{}.dt.dayofweek: {}".format(col, group_val),
            )
        elif freq == "H2":
            return (
                "{}.dt.hour == {}".format(build_col_key(col), group_val),
                "{}.dt.hour: {}".format(col, group_val),
            )
        elif freq == "H":
            ts_val = pd.Timestamp(group_val)
            day = ts_val.strftime("%Y%m%d")
            hour = ts_val.hour
            return (
                "{col}.dt.date == '{day}' and {col}.dt.hour == {hour}".format(
                    col=build_col_key(col), day=day, hour=hour
                ),
                "{col}.dt.date: {day}, {col}.dt.hour: {hour}".format(
                    col=col, day=day, hour=hour
                ),
            )
        elif freq == "D":
            ts_val = convert_date_val_to_date(group_val)
            day = ts_val.strftime("%Y%m%d")
            return (
                "{col}.dt.date == '{day}'".format(col=build_col_key(col), day=day),
                "{}.dt.date: {}".format(col, day),
            )
        elif freq == "W":
            ts_val = convert_date_val_to_date(group_val)
            return (
                "{col}.dt.year == {year} and {col}.dt.week == {week}".format(
                    col=build_col_key(col), year=ts_val.year, week=ts_val.week
                ),
                "{col}.dt.year: {year}, {col}.dt.week: {week}".format(
                    col=col, year=ts_val.year, week=ts_val.week
                ),
            )
        elif freq == "M":
            ts_val = convert_date_val_to_date(group_val)
            return (
                "{col}.dt.year == {year} and {col}.dt.month == {month}".format(
                    col=build_col_key(col), year=ts_val.year, month=ts_val.month
                ),
                "{col}.dt.year: {year}, {col}.dt.month: {month}".format(
                    col=col, year=ts_val.year, month=ts_val.month
                ),
            )
        elif freq == "Q":
            ts_val = convert_date_val_to_date(group_val)
            return (
                "{col}.dt.year == {year} and {col}.dt.quarter == {quarter}".format(
                    col=build_col_key(col), year=ts_val.year, quarter=ts_val.quarter
                ),
                "{col}.dt.year: {year}, {col}.dt.quarter: {quarter}".format(
                    col=col, year=ts_val.year, quarter=ts_val.quarter
                ),
            )
        elif freq == "Y":
            ts_val = convert_date_val_to_date(group_val)
            return (
                "{col}.dt.year == {year}".format(
                    col=build_col_key(col), year=ts_val.year
                ),
                "{}.dt.year: {}".format(col, ts_val.year),
            )
    if group_val == "NaN":
        return (
            "{col} != {col}".format(col=build_col_key(col_def)),
            "{}: NaN".format(col_def),
        )
    if group_classifier in ["I", "F", "B"]:
        return (
            "{col} == {val}".format(
                col=build_col_key(col_def), val="{}".format(group_val).replace(",", "")
            ),
            "{}: {}".format(col_def, group_val),
        )
    if group_classifier == "D":
        group_val_base = convert_date_val_to_date(group_val).strftime("%Y%m%d")
        if pandas_util.is_pandas3():
            group_val_start = (convert_date_val_to_date(group_val) - Day()).strftime(
                "%Y%m%d"
            )
            group_val_end = (convert_date_val_to_date(group_val) + Day()).strftime(
                "%Y%m%d"
            )
            return (
                "{col} > '{start}' and {col} < '{end}'".format(
                    col=build_col_key(col_def), start=group_val_start, end=group_val_end
                ),
                "{}: {}".format(col_def, group_val_base),
            )
        group_val = group_val_base
    return (
        "{col} == '{val}'".format(col=build_col_key(col_def), val=group_val),
        "{}: {}".format(col_def, group_val),
    )


def build_group_inputs_filter(df, group_inputs):
    dtypes = get_dtypes(df)

    def _group_filter(group_val):
        for gc, gv in group_val.items():
            classifier = classify_type(dtypes[gc])
            yield group_filter_handler(gc, gv, classifier)

    def _full_filter():
        for group_val in group_inputs:
            filter_vals, label_vals = [], []
            for fv, lv in _group_filter(group_val):
                filter_vals.append(fv)
                label_vals.append(lv)
            yield " and ".join(filter_vals), ", ".join(label_vals)

    full_filters, full_labels = [], []
    for ff, fl in _full_filter():
        full_filters.append(ff)
        full_labels.append(fl)
    return ("({})".format(") or (".join(full_filters)), ", ".join(full_labels))


def retrieve_chart_data(df, *args, **kwargs):
    """
    Retrieves data from a dataframe for x, y, z & group inputs complete with date frequency
    formatting (:meth:`dtale.charts.utils.date_freq_handler`) if specified

    :param df: dataframe that contains data for chart
    :type df: :class:`pandas:pandas.DataFrame`
    :param args: columns to use
    :type args: iterable of str
    :return: dataframe of data required for chart construction
    :rtype: :class:`pandas:pandas.DataFrame`
    """
    freq_handler = date_freq_handler(df)
    cols = flatten_lists([make_list(a) for a in args])
    all_code = []
    all_data = []

    if INDEX_COL in cols:
        cols = [col for col in cols if col != INDEX_COL]
        all_data.append(pd.Series(df.index, index=df.index, name="__index__"))
        all_code.append(
            "\tpd.Series(df.index, index=df.index, name='{}'),".format(INDEX_COL)
        )

    for col in cols:
        if col is not None:
            s, code = freq_handler(col)
            all_data.append(s)
            if code is not None:
                all_code.append(code)
    all_data = pd.concat(all_data, axis=1)
    all_code = ["chart_data = pd.concat(["] + all_code + ["], axis=1)"]
    if len(make_list(kwargs.get("group_val"))):
        filters, labels = build_group_inputs_filter(all_data, kwargs["group_val"])
        all_data = run_query(all_data, filters)
        all_code.append(
            "chart_data = chart_data.query({})".format(triple_quote(filters))
        )
    return all_data, all_code


def check_all_nan(df, cols=None):
    """
    Checker function to test whether all data within a column of a dataframe is :attr:`numpy:numpy.nan`

    :param df: dataframe whose data needs to be checked
    :type df: :class:`pandas:pandas.DataFrame`
    :param cols: columns to test
    :type cols: list of str
    :raises Exception: if all data within a column is :attr:`numpy:numpy.nan`
    """
    for col in cols or df.columns:
        if df[col].isnull().all():
            raise Exception('All data for column "{}" is NaN!'.format(col))


DUPES_MSG = (
    "The grouping [{}] contains duplicates, please specify group or additional filtering or select 'No Aggregation' "
    "from Aggregation drop-down."
)
LIMIT_MSG = "Dataset exceeds {} records, cannot render. Please apply filter..."


def check_exceptions(
    df,
    allow_duplicates,
    unlimited_data=False,
    data_limit=15000,
    limit_msg=LIMIT_MSG,
    dupes_msg=DUPES_MSG,
):
    """
    Checker function to test the output of any chart aggregations to see if it is one of the following:
        - too large to be rendered by web client
        - contains duplicate data points which can't be rendered (ex: multiple points for a single point on the x-axis
          of a bar chart within the same series)

    :param df: dataframe whose data needs to be checked
    :type df: :class:`pandas:pandas.DataFrame`
    :param allow_duplicates: flag to allow duplicates to be ignored (usually for scatter plots)
    :type allow_duplicates: bool
    :param data_limit: maximum rows allowed for chart rendering (default: 15,000)
    :type data_limit: int, optional
    :param limit_msg: error message template
    :type limit_msg: str, optional
    :raises Exception: if any failure condition is met
    """
    if not allow_duplicates and any(df.duplicated()):
        raise ChartBuildingError(dupes_msg.format(", ".join(df.columns)))
    if not unlimited_data and len(df) > data_limit:
        raise ChartBuildingError(limit_msg.format(data_limit))


def build_aggs(y, z=None, agg=None, extended_aggregation=[]):
    z_exists = len(make_list(z))
    agg_cols = make_list(y)
    if z_exists:
        agg_cols = make_list(z)

    aggs = {}
    if not len(extended_aggregation or []):
        aggs[agg] = agg_cols
    else:
        for ext_agg in extended_aggregation:
            aggs[ext_agg["agg"]] = aggs.get(ext_agg["agg"], []) + [ext_agg["col"]]
    return aggs


def build_agg_data(
    df,
    x,
    y,
    inputs,
    agg,
    z=None,
    group_col=None,
    animate_by=None,
    extended_aggregation=[],
    dropna=True,
):
    """
    Builds aggregated data when an aggregation (sum, mean, max, min...) is selected from the front-end.

    :param df: dataframe that contains data for chart
    :type df: :class:`pandas:pandas.DataFrame`
    :param x: column to use for the X-Axis
    :type x: str
    :param y: columns to use for the Y-Axes
    :type y: list of str
    :param inputs: additional chart configurations (chart_type, group, rolling_win, rolling_comp...)
    :type inputs: dict
    :param agg: points to a specific function that can be applied to
                :func: pandas.core.groupby.DataFrameGroupBy.  Possible values are: count, first, last mean,
                median, min, max, std, var, mad, prod, sum
    :type agg: str
    :param z: column to use for the Z-Axis
    :type z: str, optional
    :param group_col: column to use for grouping
    :type group_col: str, optional
    :param animate_by: column to use for break up data into frames for animation
    :type animate_by: str, optional
    :param extended_aggregation: list of configurations that point to a specific function that can be applied to
                :func: pandas.core.groupby.DataFrameGroupBy.  Possible values are: count, first, last mean,
                median, min, max, std, var, mad, prod, sum
    :type extended_aggregation: list, optional
    :return: dataframe of aggregated data
    :rtype: :class:`pandas:pandas.DataFrame`
    """
    z_exists = len(make_list(z))
    idx_cols = make_list(animate_by) + make_list(group_col) + [x]
    agg_cols = make_list(y)
    if z_exists:
        idx_cols += make_list(y)
        agg_cols = make_list(z)

    is_agg = not len(extended_aggregation or [])
    if is_agg and agg == "raw":
        return df, [], agg_cols

    if is_agg and agg == "corr":
        if not z_exists:
            raise NotImplementedError(
                "Correlation aggregation is only available for 3-dimensional charts!"
            )
    if is_agg and agg == "rolling":
        if z_exists:
            raise NotImplementedError(
                "Rolling computations have not been implemented for 3-dimensional charts!"
            )
        window, comp = map(inputs.get, ["rolling_win", "rolling_comp"])
        agg_df = df.set_index(x).rolling(window=window)
        agg_df = pd.DataFrame({c: getattr(agg_df[c], comp)() for c in y})
        agg_df = agg_df.reset_index()
        code = [
            "chart_data = chart_data.set_index('{x}').rolling(window={window})".format(
                x=x, window=window
            ),
            "chart_data = pd.DataFrame({'"
            + ", ".join(
                ["'{c}': chart_data['{c}'].{comp}()".format(c=c, comp=comp) for c in y]
            )
            + "})",
            "chart_data = chart_data.reset_index()",
        ]
        return agg_df, code, y

    aggs = build_aggs(y, z, agg, extended_aggregation)

    if "drop_duplicates" in aggs:
        groups = [
            df[idx_cols + [col]].drop_duplicates() for col in aggs["drop_duplicates"]
        ]
        if len(groups) == 1:
            groups = groups[0]
            code = "chart_data = chart_data[['{}']].drop_duplicates()".format(
                "','".join(idx_cols + aggs["drop_duplicates"])
            )
        else:
            groups = pd.merge(*groups, on=idx_cols, how="outer")
            code = (
                "idx_cols = ['{}']\n"
                "agg_cols = ['{}']\n"
                "chart_data = pd.merge(\n"
                "\t*[chart_data[idx_cols + [col]].drop_duplicates() for col in agg_cols],\n"
                "\ton=idx_cols,\n"
                "\thow='outer'\n"
                ")"
            ).format("','".join(idx_cols), "','".join(aggs["drop_duplicates"]))
        group_cols = [
            "{}{}drop_duplicates".format(col, CHART_JOINER_CHAR)
            for col in aggs["drop_duplicates"]
        ]
        groups.columns = idx_cols + group_cols
    else:
        groups = pandas_util.groupby(df, idx_cols, dropna=dropna)
        groups, code, group_cols = compute_aggs(
            df, groups, aggs, idx_cols, group_col, dropna=dropna
        )

    if animate_by is not None:
        full_idx = pd.MultiIndex.from_product(
            [df[c].unique() for c in idx_cols], names=idx_cols
        )
        groups = groups.reindex(full_idx).fillna(0)
        code += [
            "idx_cols = ['{cols}']".format(cols="', '".join(idx_cols)),
            "full_idx = pd.MultiIndex.from_product([df[c].unique() for c in idx_cols], names=idx_cols)"
            "chart_data = chart_data.reindex(full_idx).fillna(0)",
        ]
    if groups.index.name != "index":
        return groups.reset_index(), code, group_cols
    return groups, code, group_cols


def build_final_cols(y, z, agg, extended_aggregation):
    if not len(extended_aggregation or []):
        z = make_list(z)
        cols = y if not len(z) else z
        if agg is not None and agg != "raw":
            return ["{}{}{}".format(col, CHART_JOINER_CHAR, agg) for col in cols]
        return cols
    return [
        "{}{}{}".format(ext_agg["col"], CHART_JOINER_CHAR, ext_agg["agg"])
        for ext_agg in extended_aggregation
    ]


def parse_final_col(final_col):
    y_segs = final_col.split(CHART_JOINER_CHAR)
    if y_segs[-1] in AGGS:
        return CHART_JOINER_CHAR.join(y_segs[:-1]), y_segs[-1]
    return final_col, None


def compute_aggs(df, groups, aggs, idx_cols, group_col, dropna=True):
    all_code = []
    all_calculated_aggs = []
    all_calculated_cols = []
    for curr_agg, curr_agg_cols in aggs.items():
        chart_data_key = "chart_data_{}".format(curr_agg)
        if curr_agg in ["pctsum", "pctct"]:
            func = "sum" if curr_agg == "pctsum" else "size"
            subidx_cols = [c for c in idx_cols if c not in make_list(group_col)]
            calc_group = getattr(groups[curr_agg_cols], func)()
            calc_group = (
                calc_group
                / getattr(
                    pandas_util.groupby(df, subidx_cols, dropna=dropna)[curr_agg_cols],
                    func,
                )()
                * 100
            )
            if isinstance(calc_group, pd.Series):
                calc_group.name = curr_agg_cols[0]
                calc_group = calc_group.to_frame()

            if len(curr_agg_cols) > 1:
                groups.columns = curr_agg_cols
            elif len(curr_agg_cols) == 1:
                groups.name = curr_agg_cols[0]
            code = (
                "{chart_data} = chart_data{groupby}[['{agg_cols}']].{agg}()\n"
                "{chart_data} = {chart_data} / {chart_data}{subgroupby}.{agg}()"
            )
            code = code.format(
                groupby=pandas_util.groupby_code(idx_cols, dropna=dropna),
                subgroupby=pandas_util.groupby_code(subidx_cols, dropna=dropna),
                agg_cols="', '".join(make_list(curr_agg_cols)),
                agg=func,
                chart_data=chart_data_key,
            )
            all_code.append(code)
        elif curr_agg in ["first", "last"]:
            agg_func = "head" if curr_agg == "first" else "tail"

            def _build_first_last():
                for col in curr_agg_cols:
                    yield groups[[col]].apply(
                        lambda x: getattr(
                            x.sort_values(by=col, ascending=True), agg_func
                        )(1)
                    ).reset_index(-1, drop=True)

            calc_group = pd.concat(list(_build_first_last()), axis=1)
            all_code += [
                (
                    "groups = chart_data{groupby}\n"
                    "\ndef _build_first_last():\n"
                    "\tfor col in ['{agg_cols}']:\n"
                    "\t\tyield groups[[col]].apply(\n"
                    "\t\t\tlambda x: x.sort_values(by=col, ascending=True).{agg_func}(1)\n"
                    "\t\t)\n\n"
                    "{chart_data} = pd.DataFrame(list(_build_first_last()), columns=['{agg_cols}'])"
                ).format(
                    groupby=pandas_util.groupby_code(idx_cols, dropna=dropna),
                    agg_cols="', '".join(curr_agg_cols),
                    agg_func=agg_func,
                    chart_data=chart_data_key,
                )
            ]
        else:
            calc_group = getattr(groups[curr_agg_cols], curr_agg)()
            all_code += [
                "{chart_data} = chart_data{groupby}[['{agg_cols}']].{agg}()".format(
                    groupby=pandas_util.groupby_code(idx_cols, dropna=dropna),
                    agg_cols="', '".join(curr_agg_cols),
                    agg=curr_agg,
                    chart_data=chart_data_key,
                )
            ]
        final_cols = [
            "{}{}{}".format(col, CHART_JOINER_CHAR, curr_agg)
            for col in calc_group.columns
        ]
        all_code.append(
            "{chart_data}.columns = ['{cols}']".format(
                chart_data=chart_data_key, cols="','".join(final_cols)
            )
        )
        calc_group.columns = final_cols
        all_calculated_aggs.append(calc_group)
        all_calculated_cols += final_cols

    if len(all_calculated_aggs) > 1:
        all_code.append(
            "chart_data = pd.concat([{chart_data}], axis=1).reset_index()".format(
                chart_data=", ".join(["chart_data_{}".format(k) for k in aggs])
            )
        )
        ret_data = pd.concat(all_calculated_aggs, axis=1)
        ret_data.index.names = idx_cols
        return ret_data, all_code, all_calculated_cols
    all_code.append(
        "chart_data = chart_data_{}.reset_index()".format(list(aggs.keys())[0])
    )
    return all_calculated_aggs[0], all_code, all_calculated_cols


def build_base_chart(
    raw_data,
    x,
    y,
    group_col=None,
    group_type=None,
    group_val=None,
    bins_val=None,
    bin_type=None,
    agg=None,
    extended_aggregation=[],
    allow_duplicates=False,
    return_raw=False,
    unlimited_data=False,
    animate_by=None,
    cleaners=[],
    dropna=True,
    **kwargs
):
    """
    Helper function to return data for 'chart-data' & 'correlations-ts' endpoints.  Will return a dictionary of
    dictionaries (one for each series) which contain the data for the x & y axes of the chart as well as the minimum &
    maximum of all the series for the y-axis.  If there is only one series (no group_col specified) the only key in the
    dictionary of series data will be 'all' otherwise the keys will be the values of the groups.

    :param raw_data: dataframe to be used for chart
    :type raw_data: :class:`pandas:pandas.DataFrame`
    :param x: column to be used as x-axis of chart
    :type x: str
    :param y: column to be used as y-axis of chart
    :type y: list of strings
    :param group: comma-separated string of columns to group chart data by
    :type group: str, optional
    :param agg: points to a specific function that can be applied to
                        :func: pandas.core.groupby.DataFrameGroupBy.  Possible values are: count, first, last mean,
                        median, min, max, std, var, mad, prod, sum
    :type agg: str, optional
    :param extended_aggregation: list of configurations that point to a specific function that can be applied to
                        :func: pandas.core.groupby.DataFrameGroupBy.  Possible values are: count, first, last mean,
                        median, min, max, std, var, mad, prod, sum
    :type agg: list, optional
    :param allow_duplicates: flag to allow duplicates to be ignored (usually for scatter plots)
    :type allow_duplicates: bool, optional
    :return: dict
    """
    group_fmt_overrides = {
        "I": lambda v, nan_display, as_string: json_int(
            v, nan_display=nan_display, as_string=as_string, fmt="{}"
        )
    }
    data, code = retrieve_chart_data(
        raw_data, x, y, kwargs.get("z"), group_col, animate_by, group_val=group_val
    )
    cleaners = cleaners or []
    if len(cleaners):
        for col in data.columns:
            if classify_type(find_dtype(data[col])) == "S":
                code.append("s = chart_data['{}']".format(col))
                cleaned_col, cleaned_code = handle_cleaners(
                    data[col], ",".join(cleaners)
                )
                pandas_util.assign_col_data(data, col, cleaned_col)
                code += cleaned_code
                code.append("chart_data.loc[:, '{}'] = s".format(col))

    x_col = str("x")
    if x is None:
        x = x_col
        sequential_ints = range(1, len(data) + 1)  # sequential integers: 1, 2, ..., N
        pandas_util.assign_col_data(data, x_col, sequential_ints)
    y_cols = make_list(y)

    z_col = kwargs.get("z")
    z_cols = make_list(z_col)
    y_cols = [str(col) for col in y_cols]
    is_z = len(z_cols) > 0
    y_group_cols = y_cols if is_z else []
    sort_cols = y_group_cols
    final_cols = y_cols + z_cols
    if group_col is not None and len(group_col):
        for col in make_list(group_col):
            classifier = classify_type(find_dtype(data[col]))
            if classifier == "F" or (classifier == "I" and group_type == "bins"):
                if bin_type == "width":
                    kwargs = (
                        {"duplicates": "drop"}
                        if pandas_util.check_pandas_version("0.23.0")
                        else {}
                    )

                    pandas_util.assign_col_data(
                        data,
                        col,
                        pd.qcut(data[col], q=bins_val, **kwargs).astype("str"),
                    )
                    kwargs_str = (
                        ', duplicates="drop"'
                        if pandas_util.check_pandas_version("0.23.0")
                        else ""
                    )
                    code.append(
                        (
                            "chart_data.loc[:, '{col}'] = pd.qcut(chart_data['{col}'], q={bins}{kwargs})"
                        ).format(col=col, bins=bins_val, kwargs=kwargs_str)
                    )
                else:
                    bins_data = data[col].dropna()
                    npt = len(bins_data)
                    equal_freq_bins = np.interp(
                        np.linspace(0, npt, bins_val + 1),
                        np.arange(npt),
                        np.sort(bins_data),
                    )
                    kwargs = (
                        {"duplicates": "drop"}
                        if pandas_util.check_pandas_version("0.23.0")
                        else {}
                    )
                    pandas_util.assign_col_data(
                        data,
                        col,
                        pd.cut(data[col], bins=equal_freq_bins, **kwargs).astype("str"),
                    )
                    cut_kwargs_str = (
                        ', duplicates="drop"'
                        if pandas_util.check_pandas_version("0.23.0")
                        else ""
                    )
                    code.append(
                        (
                            "bins_data = data['{col}'].dropna()\n"
                            "npt = len(bins_data)\n"
                            "equal_freq_bins = np.interp(np.linspace(0, npt, {bins}), np.arange(npt), "
                            "np.sort(bins_data))\n"
                            "chart_data.loc[:, '{col}'] = pd.cut(chart_data['{col}'], bins=equal_freq_bins"
                            "{cut_kwargs})"
                        ).format(col=col, bins=bins_val + 1, cut_kwargs=cut_kwargs_str)
                    )

        main_group = group_col
        if animate_by is not None:
            main_group = [animate_by] + main_group
        sort_cols = main_group + [x] + sort_cols
        data = data.sort_values(sort_cols)
        code.append(
            "chart_data = chart_data.sort_values(['{cols}'])".format(
                cols="', '".join(sort_cols)
            )
        )
        nan_cols = list(data.columns)
        if not dropna:
            nan_cols = [col for col in data.columns if col not in group_col]
        check_all_nan(data, cols=nan_cols)
        data = data.rename(columns={x: x_col})
        code.append(
            "chart_data = chart_data.rename(columns={'" + x + "': '" + x_col + "'})"
        )

        if agg is not None or len(extended_aggregation):
            data, agg_code, final_cols = build_agg_data(
                data,
                x_col,
                y_cols,
                kwargs,
                agg,
                z=z_col,
                group_col=group_col,
                animate_by=animate_by,
                extended_aggregation=extended_aggregation,
                dropna=dropna,
            )
            code += agg_code

        group_vals = data[group_col].drop_duplicates()
        if len(group_vals) > MAX_GROUPS:
            dtypes = get_dtypes(group_vals)
            group_fmts = {
                c: find_dtype_formatter(dtypes[c], overrides=group_fmt_overrides)
                for c in group_col
            }

            group_f, _ = build_formatters(group_vals)
            group_vals = group_f.format_lists(group_vals)
            group_vals = pd.DataFrame(group_vals, columns=group_col)
            msg = (
                "Group ({}) contains more than {} unique values, more groups than that will make the chart unreadable. "
                'You can choose specific groups to display from then "Group(s)" dropdown above. The available group(s) '
                "are listed below:\n\n{}"
            ).format(
                ", ".join(group_col), MAX_GROUPS, group_vals.to_string(index=False)
            )
            raise ChartBuildingError(msg, group_vals.to_string(index=False))

        if dropna:
            data = data.dropna()
            code.append("chart_data = chart_data.dropna()")
        if return_raw:
            return data.rename(columns={x_col: x})
        data_f, range_f = build_formatters(data)
        ret_data = dict(
            data={},
            min={
                col: fmt(data[col].min(), None)
                for _, col, fmt in range_f.fmts
                if col in [x_col] + final_cols
            },
            max={
                col: fmt(data[col].max(), None)
                for _, col, fmt in range_f.fmts
                if col in [x_col] + final_cols
            },
        )

        dtypes = get_dtypes(data)
        group_fmts = {
            c: find_dtype_formatter(dtypes[c], overrides=group_fmt_overrides)
            for c in group_col
        }

        def _load_groups(df):
            for group_val, grp in pandas_util.groupby(df, group_col, dropna=dropna):

                def _group_filter():
                    for gv, gc in zip(make_list(group_val), group_col):
                        classifier = classify_type(dtypes[gc])
                        yield group_filter_handler(
                            gc,
                            group_fmts[gc](gv, nan_display="NaN", as_string=True),
                            classifier,
                        )

                final_group_filter, final_group_label = [], []
                for gf, gl in _group_filter():
                    final_group_filter.append(gf)
                    final_group_label.append(gl)
                group_filter = " and ".join(final_group_filter)
                group_label = "({})".format(", ".join(final_group_label))
                data = data_f.format_lists(grp)
                data["_filter_"] = group_filter
                yield group_label, data

        if animate_by is not None:
            frame_fmt = find_dtype_formatter(
                dtypes[animate_by], overrides=group_fmt_overrides
            )
            ret_data["frames"] = []
            for frame_key, frame in data.sort_values(animate_by).groupby(animate_by):
                ret_data["frames"].append(
                    dict(
                        data=dict(_load_groups(frame)),
                        name=frame_fmt(frame_key, nan_display="NaN", as_string=True),
                    )
                )
            ret_data["data"] = copy.deepcopy(ret_data["frames"][-1]["data"])
        else:
            ret_data["data"] = dict(_load_groups(data))
        return ret_data, code

    main_group = [x]
    if animate_by is not None:
        main_group = [animate_by] + main_group
    sort_cols = main_group + sort_cols
    data = data.sort_values(sort_cols)
    code.append(
        "chart_data = chart_data.sort_values(['{cols}'])".format(
            cols="', '".join(sort_cols)
        )
    )
    check_all_nan(data)
    data = data[main_group + final_cols]

    data = data.rename(columns={x: x_col})
    main_group = [x_col if c == x else c for c in main_group]
    code.append(
        "chart_data = chart_data.rename(columns={'" + x + "': '" + x_col + "'})"
    )

    # convert booleans into integers for aggregation
    for col in z_cols or y_cols:
        classifier = classify_type(find_dtype(data[col]))
        if classifier == "B":
            pandas_util.assign_col_data(data, col, data[col].astype("int"))

    if agg is not None or len(extended_aggregation):
        data, agg_code, final_cols = build_agg_data(
            data,
            x_col,
            y_cols,
            kwargs,
            agg,
            z=z_col,
            animate_by=animate_by,
            extended_aggregation=extended_aggregation,
        )
        code += agg_code
    data = data.dropna()

    if return_raw:
        return data.rename(columns={x_col: x})
    code.append("chart_data = chart_data.dropna()")

    dupe_cols = main_group + y_group_cols
    data_limit = global_state.get_chart_settings()[
        "3d_points" if is_z or animate_by is not None else "scatter_points"
    ]
    check_exceptions(
        data[dupe_cols].rename(columns={x_col: x}),
        allow_duplicates or agg in ["raw", "drop_duplicates"],
        unlimited_data=unlimited_data,
        data_limit=data_limit,
    )
    data_f, range_f = build_formatters(data)

    ret_data = dict(
        min={
            col: fmt(data[col].min(), None)
            for _, col, fmt in range_f.fmts
            if col in [x_col] + y_group_cols + final_cols
        },
        max={
            col: fmt(data[col].max(), None)
            for _, col, fmt in range_f.fmts
            if col in [x_col] + y_group_cols + final_cols
        },
    )
    if animate_by is not None:
        frame_fmt = find_dtype_formatter(
            find_dtype(data[animate_by]), overrides=group_fmt_overrides
        )
        ret_data["frames"] = []
        for frame_key, frame in data.sort_values(animate_by).groupby(animate_by):
            ret_data["frames"].append(
                dict(
                    data={str("all"): data_f.format_lists(frame)},
                    name=frame_fmt(frame_key, nan_display="NaN", as_string=True),
                )
            )
        ret_data["data"] = copy.deepcopy(ret_data["frames"][-1]["data"])
    else:
        ret_data["data"] = {str("all"): data_f.format_lists(data)}
    return ret_data, code


WEEKDAY_MAP = {
    idx: day
    for idx, day in enumerate(["Mon", "Tues", "Wed", "Thur", "Fri", "Sat", "Sun"])
}


def weekday_tick_handler(col_data, col):
    """
    Output handler for datetime data which needs to be returned as weekdays.  If the column definition ends with '|WD'
    then the integer values in 'data' will be mapped to their standard weekday test (Mon, Tues, Wed, Thur, Fri, Sat,
    Sun)

    :param col_data: iterable of values within column
    :type col_data: list
    :param col: column definition
    :type col: str
    :return: formatted column data
    :rtype: list
    """
    if col.endswith("|WD"):
        return [WEEKDAY_MAP[d] for d in col_data]
    return col_data


def find_group_vals(df, group_cols):
    group_vals, _ = retrieve_chart_data(df, group_cols)
    group_vals = group_vals.drop_duplicates().sort_values(group_cols)
    group_f, _ = build_formatters(group_vals, nan_display="NaN")
    return group_f.format_dicts(group_vals.itertuples())



# coding=utf-8
from enum import Enum
import numpy as np
import math
import scipy.stats as stats


class Component(Enum):
    """Enum containing the different Variance parts of GageRnR."""

    OPERATOR = 0
    PART = 1
    OPERATOR_BY_PART = 2
    MEASUREMENT = 3
    TOTAL = 4


ComponentNames = {
    Component.OPERATOR: "Operator",
    Component.PART: "Part",
    Component.OPERATOR_BY_PART: "Operator by Part",
    Component.MEASUREMENT: "Measurement",
    Component.TOTAL: "Total",
    "GageRnR": "Gage R&R",
}


class Result(Enum):
    """Enum containing the measurements calculated by GageRnR."""

    DF = 0
    Mean = 1
    SS = 3
    MS = 4
    Var = 5
    Std = 6
    F = 7
    P = 8
    W = 9
    K = 10
    Bias = 11


ResultNames = {
    Result.DF: "DF",
    Result.SS: "SS",
    Result.MS: "MS",
    Result.Var: "Var (σ²)",
    Result.Std: "Std (σ)",
    Result.F: "F-value",
    Result.P: "P-value",
}


class Statistics(object):
    title = "Statistics"

    def __init__(self, data, labels=None):
        self.data = data
        self.parts = data.shape[1]
        self.operators = data.shape[0]
        self.measurements = data.shape[2]
        if labels is None:
            self.labels = {}
        else:
            self.labels = labels

        if "Operator" not in self.labels:
            self.labels["Operator"] = [
                ("Operator %d" % x) for x in range(self.operators)
            ]

        if "Part" not in self.labels:
            self.labels["Part"] = [("Part %d" % x) for x in range(self.parts)]

    def calculate_mean(self):
        """Calculate Mean."""
        mu = np.array([np.mean(self.data)])

        omu = np.mean(self.data, axis=1)
        omu = np.mean(omu, axis=1)

        pmu = np.mean(self.data, axis=0)
        pmu = np.mean(pmu, axis=1)

        emu = np.mean(self.data, axis=2)
        emu = emu.reshape(self.parts * self.operators)

        return {
            Component.TOTAL: mu,
            Component.OPERATOR: omu,
            Component.PART: pmu,
            Component.MEASUREMENT: emu,
        }

    def calculate_std(self):
        std = np.array([np.std(self.data, ddof=1)])
        stdo = np.std(self.data_to_operators(), axis=1, ddof=1)
        stdp = np.std(self.data_to_parts(), axis=1, ddof=1)
        return {Component.TOTAL: std, Component.OPERATOR: stdo, Component.PART: stdp}

    def calculate(self):
        self.result = dict()
        self.result[Result.Mean] = self.calculate_mean()
        self.result[Result.Std] = self.calculate_std()

    def data_to_parts(self):
        data = np.transpose(self.data, axes=(1, 0, 2))
        return data.reshape(self.parts, self.measurements * self.operators)

    def data_to_operators(self):
        return self.data.reshape(self.operators, self.measurements * self.parts)


class GageRnR(Statistics):
    """Main class for calculating GageRnR."""

    GRR = "GageRnR"
    title = "Gage R&R"

    def __init__(self, data):
        """Initialize GageRnR algorithm.

        :param numpy.array data:
            The data tha we want to analyse using GageRnR.
            The input should be structured in a 3d array
            n[i,j,k] where i = operator, j = part, k = measurement
        """
        super(GageRnR, self).__init__(data)

    def calculate(self):
        """Calculate GageRnR."""
        self.result = dict()
        self.result[Result.DF] = self.calculate_dof()
        self.result[Result.Mean] = self.calculate_mean()
        self.result[Result.SS] = self.calculate_ss()

        self.result[Result.MS] = self.calculate_ms(
            self.result[Result.DF], self.result[Result.SS]
        )

        self.result[Result.Var] = self.calculate_var(self.result[Result.MS])

        self.result[Result.Std] = self.calculate_std(self.result[Result.Var])

        self.result[Result.F] = self.calculate_f(self.result[Result.MS])

        self.result[Result.P] = self.calculate_p(
            self.result[Result.DF], self.result[Result.F]
        )

        return self.result

    def calculate_dof(self):
        """Calculate Degrees of freedom."""
        oDoF = self.operators - 1
        pDoF = self.parts - 1
        opDoF = (self.parts - 1) * (self.operators - 1)
        eDof = self.parts * self.operators * (self.measurements - 1)
        totDof = self.parts * self.operators * self.measurements - 1
        return {
            Component.OPERATOR: oDoF,
            Component.PART: pDoF,
            Component.OPERATOR_BY_PART: opDoF,
            Component.MEASUREMENT: eDof,
            Component.TOTAL: totDof,
        }

    def calculate_squares(self):
        """Calculate Squares."""
        mean = self.calculate_mean()
        tS = (self.data - mean[Component.TOTAL]) ** 2
        oS = (mean[Component.OPERATOR] - mean[Component.TOTAL]) ** 2
        pS = (mean[Component.PART] - mean[Component.TOTAL]) ** 2

        dataE = self.data.reshape(self.operators * self.parts, self.measurements)
        meanMeas = np.repeat(mean[Component.MEASUREMENT], self.measurements)
        meanMeas = meanMeas.reshape(self.operators * self.parts, self.measurements)

        mS = (dataE - meanMeas) ** 2
        return {
            Component.TOTAL: tS,
            Component.OPERATOR: oS,
            Component.PART: pS,
            Component.MEASUREMENT: mS,
        }

    def calculate_sum_of_deviations(self):
        """Calculate Sum of Deviations."""
        squares = self.calculate_squares()
        SD = dict()
        for key in squares:
            SD[key] = np.sum(squares[key])
        return SD

    def calculate_ss(self):
        """Calculate Sum of Squares."""
        SS = self.calculate_sum_of_deviations()

        SS[Component.OPERATOR] = self.parts * self.measurements * SS[Component.OPERATOR]
        SS[Component.PART] = self.operators * self.measurements * SS[Component.PART]
        SS[Component.OPERATOR_BY_PART] = SS[Component.TOTAL] - (
            SS[Component.OPERATOR] + SS[Component.PART] + SS[Component.MEASUREMENT]
        )
        return SS

    def calculate_ms(self, dof, SS):
        """Calculate Mean of Squares."""
        MS = dict()

        for key in SS:
            MS[key] = SS[key] / dof[key]
        return MS

    def calculate_var(self, MS):
        """Calculate GageRnR Variances."""
        Var = dict()

        Var[Component.MEASUREMENT] = MS[Component.MEASUREMENT]
        Var[Component.OPERATOR_BY_PART] = (
            MS[Component.OPERATOR_BY_PART] - MS[Component.MEASUREMENT]
        ) / self.parts
        Var[Component.OPERATOR] = (
            MS[Component.OPERATOR] - MS[Component.OPERATOR_BY_PART]
        ) / (self.parts * self.measurements)
        Var[Component.PART] = (MS[Component.PART] - MS[Component.OPERATOR_BY_PART]) / (
            self.operators * self.measurements
        )

        for key in Var:
            if Var[key] < 0:
                Var[key] = 0

        Var[Component.TOTAL] = (
            Var[Component.OPERATOR]
            + Var[Component.PART]
            + Var[Component.OPERATOR_BY_PART]
            + Var[Component.MEASUREMENT]
        )

        Var[GageRnR.GRR] = (
            Var[Component.MEASUREMENT]
            + Var[Component.OPERATOR]
            + Var[Component.OPERATOR_BY_PART]
        )

        return Var

    def calculate_std(self, Var):
        """Calculate GageRnR Standard Deviations."""
        Std = dict()
        for key in Var:
            Std[key] = math.sqrt(Var[key])

        return Std

    def calculate_f(self, MS):
        """Calculate F-Values."""
        F = dict()

        F[Component.OPERATOR] = MS[Component.OPERATOR] / MS[Component.OPERATOR_BY_PART]

        F[Component.PART] = MS[Component.PART] / MS[Component.OPERATOR_BY_PART]

        F[Component.OPERATOR_BY_PART] = (
            MS[Component.OPERATOR_BY_PART] / MS[Component.MEASUREMENT]
        )

        return F

    def calculate_p(self, dof, F):
        """Calculate P-Values."""
        P = dict()

        P[Component.OPERATOR] = stats.f.sf(
            F[Component.OPERATOR],
            dof[Component.OPERATOR],
            dof[Component.OPERATOR_BY_PART],
        )

        P[Component.PART] = stats.f.sf(
            F[Component.PART], dof[Component.PART], dof[Component.OPERATOR_BY_PART]
        )

        P[Component.OPERATOR_BY_PART] = stats.f.sf(
            F[Component.OPERATOR_BY_PART],
            dof[Component.OPERATOR_BY_PART],
            dof[Component.MEASUREMENT],
        )
        return P



from flask import redirect, render_template, request, session, url_for
from functools import wraps

import dtale.global_state as global_state


def setup_auth(app):
    if not global_state.get_auth_settings()["active"]:
        return

    @app.route("/login", methods=["GET", "POST"])
    def login():
        if request.method == "POST":
            username, password = (request.form.get(p) for p in ["username", "password"])
            if not authenticate(username, password):
                return render_template(
                    "dtale/login.html", error="Invalid credentials!", page="login"
                )
            session["logged_in"] = True
            session["username"] = request.form["username"]
            return redirect(session.get("next") or "/")
        return render_template("dtale/login.html", page="login")

    @app.route("/logout")
    def logout():
        session.clear()
        return redirect(url_for("login"))


def authenticate(username, password):
    auth_settings = global_state.get_auth_settings()
    if username == auth_settings["username"] and password == auth_settings["password"]:
        return True
    return False


def requires_auth(f):
    @wraps(f)
    def decorated(*args, **kwargs):
        if not global_state.get_auth_settings()["active"]:
            return f(*args, **kwargs)
        if request.path in ["/login", "/logout"]:
            return f(*args, **kwargs)
        if request.path.startswith("/dtale/static"):
            return f(*args, **kwargs)
        if not session.get("logged_in") or not session.get("username"):
            session["next"] = request.url
            return redirect(url_for("login"))
        return f(*args, **kwargs)

    return decorated



import json
import pandas as pd

import dtale.global_state as global_state


class CombineData(object):
    def __init__(self, cfg):
        action, config, datasets = (
            cfg.get(prop) for prop in ["action", "config", "datasets"]
        )
        config = json.loads(config)
        datasets = json.loads(datasets)
        self.builder = (
            MergeBuilder(config, datasets)
            if action == "merge"
            else StackBuilder(config, datasets)
        )

    def build_data(self):
        return self.builder.build_data()

    def build_code(self):
        return "\n".join(self.builder.build_code())


def build_df(dataset, is_merge=False):
    data = global_state.get_data(dataset["dataId"])
    cols = dataset.get("columns")
    cols = list(set(cols + (dataset["index"] if is_merge else []))) if cols else None
    if cols:
        data = data[cols]
    if is_merge and dataset["index"]:
        data = data.set_index(dataset["index"])
    return data


def build_code_data(datasets, is_merge=False):
    code = ["import dtale\n"]

    def build_idx(index):
        return ".setIndex(['{}'])".format("','".join(index))

    def build_cols(columns):
        return "[['{}']]".format("','".join(columns)) if columns else ""

    for idx, dataset in enumerate(datasets, 1):
        code.append(
            "df{idx} = dtale.get_instance('{id}').data{index}{cols}".format(
                idx=idx,
                id=dataset["dataId"],
                index="" if is_merge else build_idx(dataset["index"]),
                cols=build_cols(dataset.get("columns")),
            )
        )
    return code


class MergeBuilder(object):
    def __init__(self, config, datasets):
        self.config = config
        self.datasets = datasets

    def build_data(self):
        dfs = []
        for dataset in self.datasets:
            dfs.append((build_df(dataset, is_merge=True), dataset["suffix"] or None))

        how, indicator, sort = (
            self.config.get(prop) for prop in ["how", "indicator", "sort"]
        )
        left_df, left_suffix = dfs.pop(0)
        right_df, right_suffix = dfs.pop(0)
        kwargs = dict(
            how=how,
            left_index=True,
            right_index=True,
            suffixes=[left_suffix, right_suffix],
        )
        if indicator:
            kwargs["indicator"] = "merge_1"
        if sort:
            kwargs["sort"] = sort
        final_df = left_df.merge(right_df, **kwargs)
        if len(dfs):
            for idx, (df, suffix) in enumerate(dfs, 2):
                if suffix:
                    kwargs["suffixes"] = [None, suffix]
                else:
                    kwargs.pop("suffixes", None)
                if indicator:
                    kwargs["indicator"] = "merge_{}".format(idx)
                final_df = final_df.merge(df, **kwargs)
        return final_df

    def build_code(self):
        code = build_code_data(self.datasets, is_merge=True)
        how, indicator, sort = (
            self.config.get(prop) for prop in ["how", "indicator", "sort"]
        )

        def build_merge(df1, df2, left, right, ind_id=1):
            suffixes = ""
            if left.get("suffix") or right.get("suffix"):

                def suffix_str(suffix):
                    return "'{}'".format(suffix) if suffix else "None"

                suffixes = ", suffixes=[{}, {}]".format(
                    suffix_str(left.get("suffix")), suffix_str(right.get("suffix"))
                )
            sort_param = ", sort=True" if sort else ""
            indicator_param = (
                ", indicator='merge_{}".format(ind_id) if indicator else ""
            )
            return (
                "df = {df1}.merge({df2}, how='{how}', left_index=True, right_index=True"
                "{sort}{indicator}{suffixes})"
            ).format(
                df1=df1,
                df2=df2,
                how=how,
                sort=sort_param,
                indicator=indicator_param,
                suffixes=suffixes,
            )

        code.append(build_merge("df1", "df2", self.datasets[0], self.datasets[1]))
        if len(self.datasets) > 2:
            for idx, dataset in enumerate(self.datasets[2:], 3):
                code.append(build_merge("df", "df{}".format(idx), {}, dataset, idx - 1))
        return code


class StackBuilder(object):
    def __init__(self, config, datasets):
        self.config = config
        self.datasets = datasets

    def build_data(self):
        ignore_index = self.config.get("ignoreIndex")
        kwargs = {}
        if ignore_index:
            kwargs["ignore_index"] = ignore_index
        return pd.concat([build_df(dataset) for dataset in self.datasets], **kwargs)

    def build_code(self):
        code = build_code_data(self.datasets, is_merge=True)
        ignore_index = self.config.get("ignoreIndex")
        ignore_index_param = ", ignore_index=True" if ignore_index else ""
        code.append(
            "df = pd.concat([{dfs}]{ignore_index})".format(
                dfs=",".join(
                    ["df{}".format(idx) for idx, _ in enumerate(self.datasets, 1)]
                ),
                ignore_index=ignore_index_param,
            )
        )
        return code
